# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the best-batch sampler."""

import numpy as np
from numpy.typing import NDArray
from scipy.stats import betabinom

from black_it.samplers.base import BaseSampler
from black_it.search_space import SearchSpace


class BestBatchSampler(BaseSampler):
    """This class implements the best-batch sampler."""

    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample from the search space using a genetic algorithm.

        Args:
            batch_size: the number of points to sample
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the sampled parameters (an array of shape `(self.batch_size, search_space.dims)`)
        """
        if len(existing_points) < batch_size:
            raise ValueError(
                "best-batch sampler requires a number of existing points "
                f"which is at least the batch size {batch_size}, "
                f"got {len(existing_points)}"
            )

        # sort existing params
        candidate_points: NDArray[np.float64] = existing_points[
            np.argsort(existing_losses)
        ][:batch_size, :]

        candidate_point_indexes: NDArray[np.int64] = self.random_generator.integers(
            0, batch_size, size=batch_size
        )
        sampled_points: NDArray[np.float64] = np.copy(
            candidate_points[candidate_point_indexes]
        )

        beta_binom_rv = betabinom(n=search_space.dims - 1, a=3.0, b=1.0)
        beta_binom_rv.random_state = self.random_generator

        for sampled_point in sampled_points:
            num_shocks: NDArray[np.int64] = beta_binom_rv.rvs(size=1) + 1
            params_shocked: NDArray[np.int64] = self.random_generator.choice(
                search_space.dims, tuple(num_shocks), replace=False
            )
            for index in params_shocked:
                shock_size: int = self.random_generator.integers(1, 6)
                shock_sign: int = (self.random_generator.integers(0, 2) * 2) - 1

                delta: float = search_space.parameters_precision[index]
                shift: float = delta * shock_sign * shock_size
                sampled_point[index] += shift

                sampled_point[index] = np.clip(
                    sampled_point[index],
                    search_space.parameters_bounds[0][index],
                    search_space.parameters_bounds[1][index],
                )

        return sampled_points



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module defines the 'BaseLoss' base class."""

from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from black_it.utils.base import assert_


class BaseLoss(ABC):
    """BaseLoss interface."""

    def __init__(self, coordinate_weights: Optional[NDArray] = None):
        """
        Initialize the loss function.

        Args:
            coordinate_weights: the weights of the loss coordinates.
        """
        self.coordinate_weights = coordinate_weights

    def compute_loss(
        self, sim_data_ensemble: NDArray[np.float64], real_data: NDArray[np.float64]
    ) -> float:
        """
        Compute the loss between simulated and real data.

        Args:
            sim_data_ensemble: an ensemble of simulated data, of shape (ensemble_size, N, D)
            real_data: the real data, of shape (N, D)

        Returns:
            The loss value.
        """
        num_coords = real_data.shape[1]

        if self.coordinate_weights is None:
            weights = np.ones(num_coords) / num_coords
        else:
            nb_coordinate_weights = len(self.coordinate_weights)
            assert_(
                nb_coordinate_weights == num_coords,
                (
                    "the length of coordinate_weights should be equal "
                    f"to the number of coordinates, got {nb_coordinate_weights} and {num_coords}"
                ),
                exc_cls=ValueError,
            )
            weights = self.coordinate_weights

        loss = 0
        for i in range(num_coords):
            loss += (
                self.compute_loss_1d(sim_data_ensemble[:, :, i], real_data[:, i])
                * weights[i]
            )

        return loss

    @abstractmethod
    def compute_loss_1d(
        self, sim_data_ensemble: NDArray[np.float64], real_data: NDArray[np.float64]
    ) -> float:
        """
        Return the loss between a specific coordinate of two time series.

        Concrete classes have to override this method in order to implement new
        loss functions.

        Args:
            sim_data_ensemble: an ensemble of simulated 1D series, shape (ensemble_size, N, 1)
            real_data: the real data, shape (N, 1)

        Returns:
            the computed loss over the specific coordinate
        """



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the GSL-div implementation.

- Lamperti, F. (2018).
  An information theoretic criterion for empirical validation of simulation models.
  Econometrics and Statistics, 5, 83-106.

Algorithm description:

0. We have 2 TS:
    a. {X} = 'observed', 1 realisation
    b. {y} = simulated, many realisations
1. Prepare the TS (discretize) {X} = {x1,x2,...,xt,...,xT}:
    1. take {xMax,xMin}
    2. partition in 'b' intervals - simbolize the TS
        1. each 'xt' becomes the bucket # that is assigned to 'xt_'
    3. the symbolized TS {x_} is subdivided in 'words' of len 'l'
        2. use a set of len(s) for 'l' eg {1,2,...,L}
    4. for each word len 'l' compute the vector of word occurrencies
        3. hence get 'L' such frequencies vectors: {x_l}
2. Compute the GSL-div:
    1. for each word len 'l' compute:
        1. Shannon entropy of Sx = {x_l}
        2. Shannon entropy of Sm = ({x_l} + {y_l})/2,
            where {y} is another TS. Ie we obs {X} and simulate {y}.
        3. compute weights to be given to each word len 'l': {w}
            1. use 'additively progressive weights'
            2. w(0) = 0
            3. w(l+1) = w(l) + 2/(L*(L-1))
        4. GSL-div = SUMMATION(1,...,L)w*(2Sm - Sx)
                    = SUMMATION(1,...,L)w*(2(avg entropy) - obs entropy)
3. Ensable runs (k simulations of the model)
    1. use a simple average for freq vectors {x_l}


**Note**: b,L don't increase much the comp power required (ie from (2,2) to (19,19) +20% time).

"""
from typing import Optional, Union

import numpy as np
from numpy.typing import NDArray

from black_it.loss_functions.base import BaseLoss
from black_it.utils.base import assert_

EPS = 0.00001  # np.finfo(float).eps


class GslDivLoss(BaseLoss):
    """
    Class for the Gsl-div loss.

    Example:
        >>> expected_loss = 0.39737637181336855
        >>> np.random.seed(11)
        >>> series1 = np.random.normal(0, 1, (100, 3))
        >>> series2 = np.random.normal(0, 1, (100, 3))
        >>> loss_func = GslDivLoss()
        >>> loss = loss_func.compute_loss(series1[None, :, :], series2)
        >>> assert np.isclose(expected_loss, loss)
    """

    def __init__(
        self,
        nb_values: int = None,
        nb_word_lengths: int = None,
        coordinate_weights: Optional[NDArray] = None,
    ) -> None:
        """
        Initialize the GSL-div loss object.

        Args:
            nb_values: number of values the digitised series can take
            nb_word_lengths: the number of word length to consider
            coordinate_weights: the weights of the loss coordinates
        """
        super().__init__(coordinate_weights)
        self.nb_values = nb_values
        self.nb_word_lengths = nb_word_lengths

    def compute_loss_1d(
        self, sim_data_ensemble: NDArray[np.float64], real_data: NDArray[np.float64]
    ) -> float:
        """
        Return the GSL-div measure.

        From (Lamperti, 2017):

        > The information loss about the behaviour of the stochastic process
          due to the symbolization becomes smaller and smaller as b increases.
          On the other side, low values of b would likely wash away
          processes’ noise the modeller might not be interested in.

        Args:
            sim_data_ensemble: the ensemble of simulated data
            real_data: the real data

        Returns:
            the GSL loss
        """
        N = len(real_data)
        ensemble_size = sim_data_ensemble.shape[0]

        if self.nb_values is None:
            nb_values = int((N - 1) / 2.0)
        else:
            nb_values = self.nb_values

        if self.nb_word_lengths is None:
            nb_word_lengths = int((N - 1) / 2.0)
        else:
            nb_word_lengths = self.nb_word_lengths

        # discretize real time series
        obs_xd = self.discretize(
            real_data,
            nb_values,
            np.min(real_data),
            np.max(real_data),
        )

        gsl_loss = 0.0

        # average loss over the ensemble
        for sim_data in sim_data_ensemble:

            # discretize simulated series
            sim_xd = self.discretize(
                sim_data, nb_values, np.min(sim_data), np.max(sim_data)
            )

            loss = self.gsl_div_1d_1_sample(
                sim_xd, obs_xd, nb_word_lengths, nb_values, N
            )

            gsl_loss += loss

        return gsl_loss / ensemble_size

    def gsl_div_1d_1_sample(
        self,
        sim_xd: NDArray,
        obs_xd: NDArray,
        nb_word_lengths: int,
        nb_values: int,
        N: int,
    ) -> float:
        """Compute the GSL-div for a single realisation of the simulated data.

        Args:
            sim_xd: discretised simulated series
            obs_xd: discretised real series
            nb_word_lengths: the number of word length to consider
            nb_values: number of values the digitised series can take
            N: the length of real and simulated series

        Returns:
            the computed loss
        """
        # outcome measure
        gsl_div = 0.0

        # weight
        weight = 0.0

        # for any word len:
        for word_length in range(1, nb_word_lengths + 1):
            sim_xw = self.get_words(sim_xd, word_length)
            obs_xw = self.get_words(obs_xd, word_length)
            m_xw = np.concatenate((sim_xw, obs_xw))
            sim_xp = self.get_words_est_prob(sim_xw)
            m_xp = self.get_words_est_prob(m_xw)
            base = float(nb_values**word_length)
            sim_entr = self.get_sh_entr(sim_xp, base)
            m_entr = self.get_sh_entr(m_xp, base)

            # update weight
            weight = weight + 2 / (nb_word_lengths * (nb_word_lengths + 1))

            # correction
            corr = ((len(m_xp) - 1) - (len(sim_xp) - 1)) / (2 * N)

            # add to measure
            gsl_divl = 2 * m_entr - sim_entr + corr
            gsl_div = gsl_div + weight * gsl_divl

        # end of cycle, return
        return gsl_div

    @staticmethod
    def discretize(
        time_series: NDArray[np.float64],
        nb_values: int,
        start_index: Union[np.float64, float],
        stop_index: Union[np.float64, float],
    ) -> NDArray[np.int64]:
        """
        Discretize the TS in 'nb_values' finite states.

        >>> GslDivLoss.discretize(
        ...     [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        ...     nb_values=3,
        ...     start_index=1,
        ...     stop_index=10
        ... )
        array([1, 1, 1, 2, 2, 2, 2, 3, 3, 3])

        Args:
            time_series: any univariate time series
            nb_values: int, number of values the digitised series can take.
                It must be greater than 0.
            start_index: the starting point
            stop_index: the stopping point

        Returns:
            the discretised time series
        """
        start_index = np.float64(start_index)
        stop_index = np.float64(stop_index)
        linspace = np.linspace(start_index - EPS, stop_index + EPS, nb_values + 1)

        return np.searchsorted(linspace, time_series, side="left")

    @staticmethod
    def get_words(time_series: NDArray[np.float64], length: int) -> NDArray:
        """
        Return an overlapping array of words (int32) of 'length' given a discretised vector.

        >>> GslDivLoss.get_words(np.asarray([1, 2, 2, 2]), 2)
        array([12, 22, 22])

        Args:
            time_series: any univariate discretised time series
            length: int, len of words to be returned. Must be such that
                (len(time_series) + 1 - length) is positive.

        Returns:
            the time series of overlapping words

        """
        tswlen = len(time_series) + 1 - length
        assert_(tswlen >= 0, "the chosen word length is too high", exc_cls=ValueError)
        tsw = np.zeros(shape=(tswlen,), dtype=np.int32)

        for i in range(length):
            k = 10 ** (length - i - 1)
            tsw = tsw + time_series[i : tswlen + i] * k

        return tsw

    @staticmethod
    def get_words_est_prob(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
        """
        Return an array of estimated probabilities given an array of words (int32).

        Args:
            time_series: any univariate array of words

        Returns:
            estimate of probabilities
        """
        _, count = np.unique(time_series, return_counts=True)
        est_p = np.divide(count, np.sum(count))
        return est_p

    @staticmethod
    def get_sh_entr(probs: NDArray[np.float64], log_base: float) -> float:
        """
        Return the Shannon entropy given an array of probabilities.

        Args:
            probs: an array of probabilities describing the discrete probability distribution
            log_base: the entropy logarithm base.

        Returns:
            the entropy of the discrete probability distribution
        """
        log = np.log(probs) / np.log(log_base)
        return -np.sum(np.multiply(probs, log))



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of the particle swarm sampler."""
from typing import Optional, cast

import numpy as np
from numpy.typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.search_space import SearchSpace
from black_it.utils.base import assert_, digitize_data, positive_float


class ParticleSwarmSampler(
    BaseSampler
):  # pylint: disable=(too-many-instance-attributes)
    """
    Implementation of a particle swarm sampler.

    This sampler implements the particle swarm sampling method commonly used in particle swarm optimization (PSO),
    introduced in:

        Eberhart, Russell, and James Kennedy. "A new optimizer using particle swarm theory."
          MHS'95. Proceedings of the sixth international symposium on micro machine and human science. IEEE, 1995.

    In a particle swarm optimizer, there is a set of particles that are "evolved" by cooperation and competition
    among the individuals themselves through generations. Each particle adjusts its flying  according to its own
    flying experience and its companions’ flying experience. Each particle, in fact, represents a potential solution
    to a problem. Each particle is treated as a point in a D-dimensional space.  The ith particle is represented as
    Xi = (x_{i1},...,x_{iD}). The best previous position (the position giving the best fitness value) of any particle
    is recorded and represented as Pi = (p_{i1},...,p_{iD}). The index of the best particle among all the particles
    in the population is represented by the symbol g. The rate of the position change (velocity) for particle i is
    represented as Vi = (v_{i1},...,v_{iD}). The particles are manipulated according to the following equation:

        v_{id} = (ω * v_{id}) + (c1 * r1 * (p_{id} - x_{id})) + (c2 * r2 * (p_{gd} - x_{id})
        x_{id} = x_{id} + v_{id}

    Where:
        - ω is the inertia weight  to control the influence of the previous velocity;
        - c1 and c2 are positive values that represent the acceleration constants;
        - r1 and r2 are two random numbers uniformly distributed in the range of (0, 1).

    Note that p_{gd}, the global best position found across the dynamics, can optionally be computed by also
    considering the sampling performed by other samplers in order to let them interfere constructively with the
    Particle Swarm Sampler.
    """

    def __init__(
        self,
        batch_size: int,
        random_state: Optional[int] = None,
        inertia: float = 0.9,
        c1: float = 0.1,
        c2: float = 0.1,
        global_minimum_across_samplers: bool = False,
    ) -> None:
        """
        Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the random state of the sampler, fixing this number the sampler behaves deterministically
            inertia: the inertia of the particles' motion
            c1: first acceleration constant
            c2: second acceleration constant
            global_minimum_across_samplers: if True, the global minimum attractor of the particles' dynamics is computed
                taking into consideration also parameters sampled by other samplers, default is False
        """
        # max_duplication_passes must be zero because the sampler is stateful
        super().__init__(
            batch_size, random_state=random_state, max_duplication_passes=0
        )

        # The batch size is the number of sampled parameters per iteration. In a Black-it sampler, each call to
        # sample_batch represent an iteration of the particle swarm sampler, so it seems natural to set the number of
        # particles to the batch size, as at each iteration sample_batch returns the current positions of the
        # particles.
        self.nb_particles = batch_size

        self._inertia = positive_float(inertia)
        self._c1 = positive_float(c1)
        self._c2 = positive_float(c2)
        self._global_minimum_across_samplers = global_minimum_across_samplers

        # all current particle positions; shape=(nb_particles, space dimensions)
        self._curr_particle_positions: Optional[NDArray] = None
        # all current particle velocities; shape=(nb_particles, space dimensions)
        self._curr_particle_velocities: Optional[NDArray] = None
        # best particle positions, i.e. ; shape=(nb_particles, space dimensions)
        self._best_particle_positions: Optional[NDArray] = None
        # losses of the best positions
        self._best_position_losses: Optional[NDArray] = None
        # particle id of the global best particle position
        self._global_best_particle_id: Optional[int] = None

        # best point in parameter space - could be the best across samplers
        self._best_point: Optional[NDArray] = None

        self._previous_batch_index_start: Optional[int] = None

    @property
    def is_set_up(self) -> bool:
        """Return true iff the sampler is already set up."""
        return self._curr_particle_positions is not None

    @property
    def inertia(self) -> float:
        """Get the inertia weight."""
        return self._inertia

    @property
    def c1(self) -> float:
        """Get the c1 constant."""
        return self._c1

    @property
    def c2(self) -> float:
        """Get the c2 constant."""
        return self._c2

    def _set_up(self, dims: int) -> None:
        """Set up the sampler."""
        self._curr_particle_positions = self.random_generator.random(
            size=(self.batch_size, dims)
        )
        self._curr_particle_velocities = (
            self.random_generator.random(
                size=cast(NDArray, self._curr_particle_positions).shape
            )
            - 0.5
        )
        self._best_particle_positions = self._curr_particle_positions
        # set losses to inf as we are interested to the min
        self._best_position_losses = np.full(self.nb_particles, np.inf)
        # we don't know yet which is the best index - initialize to 0
        self._global_best_particle_id = 0

    def _get_best_position(self) -> NDArray[np.float64]:
        """
        Get the position corresponding to the global optimum the particles should converge to.

        If _global_minimum_across_samplers is False, then this method returns the current position
        of the particle that in its history has sampled, so far, the best set of parameters.

        Else, if _global_minimum_across_samplers is True, then this method returns the point
        in parameter space that achieved the minimum loss. Note that this point could have been
        sampled by a different sampler than "self".

        Returns:
            a Numpy array
        """
        if not self._global_minimum_across_samplers:
            best_particle_positions = cast(NDArray, self._best_particle_positions)
            return best_particle_positions[self._global_best_particle_id]
        return cast(NDArray, self._best_point)

    def reset(self) -> None:
        """Reset the sampler."""
        self._curr_particle_positions = None
        self._curr_particle_velocities = None
        self._best_particle_positions = None
        self._best_position_losses = None
        self._global_best_particle_id = None
        self._previous_batch_index_start = None
        assert_(
            not self.is_set_up,
            message="reset call did not work, sampler still set up",
            exc_cls=RuntimeError,
        )

    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Sample a batch of parameters."""
        if not self.is_set_up:
            self._set_up(search_space.dims)
            self._previous_batch_index_start = len(existing_points)
            return digitize_data(
                cast(NDArray[np.float64], self._best_particle_positions),
                search_space.param_grid,
            )

        self._update_best(existing_points, existing_losses)
        self._do_step()

        p_bounds: NDArray[np.float64] = search_space.parameters_bounds
        sampled_points = p_bounds[0] + self._curr_particle_positions * (
            p_bounds[1] - p_bounds[0]
        )
        self._previous_batch_index_start = len(existing_points)

        return digitize_data(sampled_points, search_space.param_grid)

    def _update_best(
        self, existing_points: NDArray[np.float64], existing_losses: NDArray[np.float64]
    ) -> None:
        """Update the best local and global positions."""
        assert_(
            self._previous_batch_index_start is not None,
            exc_cls=AssertionError,
            message="should have been set",
        )

        # set best loss and best point
        best_point_index = np.argmin(existing_losses)
        self._best_point = existing_points[best_point_index]

        # set best particle position
        batch_index_start = cast(int, self._previous_batch_index_start)
        batch_index_stop = batch_index_start + self.batch_size
        previous_points = existing_points[batch_index_start:batch_index_stop]
        previous_losses = existing_losses[batch_index_start:batch_index_stop]
        for particle_id, (point, loss) in enumerate(
            zip(previous_points, previous_losses)
        ):
            best_particle_positions = cast(NDArray, self._best_particle_positions)
            best_position_losses = cast(NDArray, self._best_position_losses)
            if best_position_losses[particle_id] > loss:
                best_particle_positions[particle_id] = point
                best_position_losses[particle_id] = loss

                # check if also the global best should be updated
                best_global_loss = best_position_losses[self._global_best_particle_id]
                if loss < best_global_loss:
                    self._global_best_particle_id = particle_id

    def _do_step(self) -> None:
        """Do a step by updating particle positions and velocities."""
        curr_particle_positions = cast(NDArray, self._curr_particle_positions)
        curr_particle_velocities = cast(NDArray, self._curr_particle_velocities)
        best_particle_positions = cast(NDArray, self._best_particle_positions)
        r1_vec = self.random_generator.random(size=curr_particle_positions.shape)
        r2_vec = self.random_generator.random(size=curr_particle_positions.shape)
        new_particle_velocities = (
            self.inertia * curr_particle_velocities
            + self.c1
            * r1_vec
            * (best_particle_positions - self._curr_particle_positions)
            + self.c2
            * r2_vec
            * (self._get_best_position() - self._curr_particle_positions)  # type: ignore
        )

        self._curr_particle_positions = np.clip(
            self._curr_particle_positions + new_particle_velocities,
            a_min=0.0,
            a_max=1.0,
        )
        self._curr_particle_velocities = new_particle_velocities



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the random forest sampling."""
from typing import Optional, Tuple

import numpy as np
from numpy.typing import NDArray
from sklearn.ensemble import RandomForestClassifier

from black_it.samplers.base import BaseSampler
from black_it.samplers.random_uniform import RandomUniformSampler
from black_it.search_space import SearchSpace
from black_it.utils.base import digitize_data


class RandomForestSampler(BaseSampler):
    """This class implements random forest sampling."""

    def __init__(
        self,
        batch_size: int,
        random_state: Optional[int] = None,
        max_deduplication_passes: int = 5,
        candidate_pool_size: Optional[int] = None,
        n_estimators: int = 500,
        criterion: str = "gini",
    ) -> None:
        """
        Random forest sampling.

        Note: this class is a wrapper of sklearn.ensemble.RandomForestClassifier.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the random state of the sampler, fixing this number the sampler behaves deterministically
            max_deduplication_passes: the maximum number of deduplication passes
            candidate_pool_size: number of randomly sampled points on which the random forest predictions are evaluated
            n_estimators: number of trees in the forest
            criterion: The function to measure the quality of a split.
        """
        super().__init__(batch_size, random_state, max_deduplication_passes)

        self._n_estimators = n_estimators
        self._criterion = criterion

        if candidate_pool_size is not None:
            self._candidate_pool_size = candidate_pool_size
        else:
            self._candidate_pool_size = 1000 * batch_size

    @property
    def n_estimators(self) -> int:
        """Get the number of estimators."""
        return self._n_estimators

    @property
    def criterion(self) -> str:
        """Get the criterion."""
        return self._criterion

    @property
    def candidate_pool_size(self) -> int:
        """Get the candidate pool size."""
        return self._candidate_pool_size

    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample from the search space.

        Args:
            batch_size: the number of points to sample
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the sampled parameters
        """
        # Get large candidate pool
        candidates = RandomUniformSampler(
            self.candidate_pool_size, random_state=self._get_random_seed()
        ).sample_batch(
            self.candidate_pool_size, search_space, existing_points, existing_losses
        )

        # Train surrogate
        x: NDArray[np.float64]
        y: NDArray[np.int64]
        x, y, _existing_points_quantiles = self.prepare_data_for_classifier(
            existing_points, existing_losses
        )

        classifier: RandomForestClassifier = RandomForestClassifier(
            n_estimators=self.n_estimators,
            criterion=self.criterion,
            n_jobs=-1,
            random_state=self._get_random_seed(),
        )
        classifier.fit(x, y)
        # Predict quantiles
        predicted_points_quantiles: NDArray[np.float64] = classifier.predict(candidates)
        # Sort params by predicted quantile
        sorting_indices: NDArray[np.int64] = np.argsort(predicted_points_quantiles)
        sampled_points: NDArray[np.float64] = candidates[sorting_indices][:batch_size]

        return digitize_data(sampled_points, search_space.param_grid)

    @staticmethod
    def prepare_data_for_classifier(
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
        num_bins: int = 10,
    ) -> Tuple[NDArray[np.float64], NDArray[np.int64], NDArray[np.float64]]:
        """
        Prepare data for the classifier.

        Args:
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters
            num_bins: the number of bins

        Returns:
            A triple (x, y, quantiles), where
                - x is the vector of training data
                - y is the vector of targets
                - the quantiles
        """
        x: NDArray[np.float64] = existing_points
        y: NDArray[np.float64] = existing_losses

        cutoffs: NDArray[np.float64] = np.linspace(0, 1, num_bins + 1)
        quantiles: NDArray[np.float64] = np.zeros(num_bins + 1)

        for i in range(num_bins - 1):
            quantiles[i + 1] = np.quantile(y, cutoffs[i + 1])

        quantiles[-1] = np.max(y)

        y_cat: NDArray[np.int64] = np.digitize(y, quantiles, right=True)
        y_cat = y_cat - 1

        return x, y_cat, quantiles



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains utility functions to deal with time series."""

import numpy as np
import statsmodels.api as sm
from numpy.typing import NDArray
from scipy.stats import kurtosis, skew


def get_mom_ts(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """
    Compute specific moments from a time series.

    Args:
        time_series: the time series

    Returns:
        the moments
    """
    moments = np.array([get_mom_ts_1d(ts_d) for ts_d in time_series.T])
    return moments.T


def get_mom_ts_1d(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """Compute specific moments from a time series.

    Args:
        time_series: the time series

    Returns:
        the moments from a time series
    """
    # array of all moments (to be filled)
    avg_vec_mom = np.zeros(18)

    # first 4 moments and auto-correlations of the time series
    avg_vec_mom[0] = np.mean(time_series)
    avg_vec_mom[1] = np.std(time_series)
    s, k = skew(time_series), kurtosis(time_series)  # pylint: disable=invalid-name
    avg_vec_mom[2] = np.sign(s) * np.power(abs(s), 1.0 / 3.0)
    avg_vec_mom[3] = np.sign(k) * np.power(abs(k), 1.0 / 4.0)

    ts_acf = sm.tsa.acf(time_series, nlags=5, fft=False)
    avg_vec_mom[4] = ts_acf[1]
    avg_vec_mom[5] = ts_acf[2]
    avg_vec_mom[6] = ts_acf[3]
    avg_vec_mom[7] = ts_acf[4]
    avg_vec_mom[8] = ts_acf[5]

    # first 4 moments and auto-correlations of the absolute differences of the time series
    abs_diff = np.absolute(np.diff(time_series))

    avg_vec_mom[9] = np.mean(abs_diff)
    avg_vec_mom[10] = np.std(abs_diff)
    s, k = skew(abs_diff), kurtosis(abs_diff)  # pylint: disable=invalid-name
    avg_vec_mom[11] = np.sign(s) * np.power(abs(s), 1.0 / 3.0)
    avg_vec_mom[12] = np.sign(k) * np.power(abs(k), 1.0 / 4.0)

    ts_diff_acf = sm.tsa.acf(abs_diff, nlags=5, fft=False)
    avg_vec_mom[13] = ts_diff_acf[1]
    avg_vec_mom[14] = ts_diff_acf[2]
    avg_vec_mom[15] = ts_diff_acf[3]
    avg_vec_mom[16] = ts_diff_acf[4]
    avg_vec_mom[17] = ts_diff_acf[5]

    np.nan_to_num(avg_vec_mom, False)

    return avg_vec_mom



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module defines the 'BaseSampler' base class."""

from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np
from numpy.random import default_rng
from numpy.typing import NDArray

from black_it.search_space import SearchSpace
from black_it.utils.base import get_random_seed


class BaseSampler(ABC):
    """
    BaseSampler interface.

    This is the base class for all samplers.
    """

    def __init__(
        self,
        batch_size: int,
        random_state: Optional[int] = None,
        max_duplication_passes: int = 5,
    ) -> None:
        """
        Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the internal state of the sampler, fixing this numbers the sampler behaves deterministically
            max_duplication_passes: maximum number of duplication passes done to avoid sampling repeated parameters
        """
        self.random_state: Optional[int] = random_state
        self.batch_size: int = batch_size
        self.max_duplication_passes = max_duplication_passes

    @property
    def random_state(self) -> Optional[int]:
        """Get the random state."""
        return self._random_state

    @random_state.setter
    def random_state(self, random_state: Optional[int]) -> None:
        """Set the random state."""
        self._random_state = random_state
        self._random_generator = default_rng(self.random_state)

    @property
    def random_generator(self) -> np.random.Generator:
        """Get the random generator."""
        return self._random_generator

    def _get_random_seed(self) -> int:
        """Get new random seed from the current random generator."""
        return get_random_seed(self._random_generator)

    @abstractmethod
    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample a number of new parameters fixed by the 'batch_size' attribute.

        Args:
            batch_size: number of samples to collect
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the new parameters
        """

    def sample(
        self,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample from the search space.

        Args:
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the sampled parameters
        """
        samples = self.sample_batch(
            self.batch_size, search_space, existing_points, existing_losses
        )

        for n in range(self.max_duplication_passes):

            duplicates = self.find_and_get_duplicates(samples, existing_points)

            num_duplicates = len(duplicates)

            if num_duplicates == 0:
                break

            new_samples = self.sample_batch(
                num_duplicates, search_space, existing_points, existing_losses
            )
            samples[duplicates] = new_samples

            if n == self.max_duplication_passes - 1:
                print(
                    f"Warning: Repeated samples still found after {self.max_duplication_passes} duplication passes."
                    " This is probably due to a small search space."
                )

        return samples

    @staticmethod
    def find_and_get_duplicates(
        new_points: NDArray[np.float64], existing_points: NDArray[np.float64]
    ) -> List:
        """Find the points in 'new_points' that are already present in 'existing_points'.

        Args:
            new_points: candidates points for the sampler
            existing_points: previously sampled points

        Returns:
            the location of the duplicates in 'new_points'
        """
        all_points = np.concatenate((existing_points, new_points))
        unq, count = np.unique(all_points, axis=0, return_counts=True)
        repeated_groups = unq[count > 1]

        repeated_pos = []
        if len(repeated_groups) > 0:
            for repeated_group in repeated_groups:
                repeated_idx = np.argwhere(np.all(new_points == repeated_group, axis=1))
                for index in repeated_idx:
                    repeated_pos.append(index[0])

        return repeated_pos



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2022 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation for the Halton sampler."""
import itertools
from typing import Iterator, List, Optional

import numpy as np
from numpy.random import default_rng
from numpy.typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.search_space import SearchSpace
from black_it.utils.base import check_arg, digitize_data

_MIN_SEQUENCE_START_INDEX = 20
_MAX_SEQUENCE_START_INDEX = 2**16


class HaltonSampler(BaseSampler):
    """
    Halton low discrepancy sequence.

    This snippet implements the Halton sequence following the generalization of
    a sequence of *Van der Corput* in n-dimensions.
    """

    def __init__(
        self,
        batch_size: int,
        random_state: Optional[int] = None,
        max_deduplication_passes: int = 5,
    ) -> None:
        """
        Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the random state of the sampler, fixing this number the sampler behaves deterministically
            max_deduplication_passes: the maximum number of sample deduplication passes.
        """
        super().__init__(batch_size, random_state, max_deduplication_passes)
        self._prime_number_generator = _CachedPrimesCalculator()

        # drop first N entries to avoid linear correlation
        self._reset_sequence_index()

    @property
    def random_state(self) -> Optional[int]:
        """Get the random state."""
        return self._random_state

    @random_state.setter
    def random_state(self, random_state: Optional[int]) -> None:
        """Set the random state."""
        self._random_state = random_state
        self._random_generator = default_rng(self.random_state)
        self._reset_sequence_index()

    def _reset_sequence_index(self) -> None:
        """Reset the sequence index pointer."""
        self._sequence_index = self.random_generator.integers(
            _MIN_SEQUENCE_START_INDEX, _MAX_SEQUENCE_START_INDEX
        )

    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """
        Sample points using Halton sequence.

        Args:
            batch_size: the number of samples
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled (not used)
            existing_losses: the loss corresponding to the sampled parameters (not used)

        Returns:
            the parameter sampled
        """
        unit_cube_points: NDArray[np.float64] = self._halton(
            batch_size, search_space.dims
        )
        p_bounds: NDArray[np.float64] = search_space.parameters_bounds
        sampled_points = p_bounds[0] + unit_cube_points * (p_bounds[1] - p_bounds[0])
        return digitize_data(sampled_points, search_space.param_grid)

    def _halton(self, nb_samples: int, dims: int) -> NDArray[np.float64]:
        """
        Get a Halton sequence.

        It uses a simple prime number generator, which takes the first `dims` primes.

        Args:
            nb_samples: number of samples
            dims: the number of dimensions of the space to sample from the unitary cube

        Returns:
            sequence of Halton.
        """
        bases: NDArray[np.int64] = self._prime_number_generator.get_n_primes(dims)
        # Generate a sample using a Halton sequence.
        sample: NDArray[np.float64] = halton(
            sample_size=nb_samples, bases=bases, n_start=self._sequence_index
        )

        # increment sequence start index for the next sampling
        self._sequence_index += nb_samples
        return sample


class _PrimesIterator:
    """
    This class implements an iterator that iterates over all primes via unbounded Sieve of Erathosthenes.

    Adapted from:

        https://wthwdik.wordpress.com/2007/08/30/an-unbounded-sieve-of-eratosthenes/

    It caches the sequence of primes up to the highest n.
    """

    def __init__(self) -> None:
        """Initialize the iterator."""
        self._primes = [[2, 2]]
        self._candidate = 2

    def __iter__(self) -> Iterator:
        """Make the class iterable."""
        return self

    def __next__(self) -> int:
        """Get the next prime number."""
        while True:
            self._candidate = self._candidate + 1
            for i in self._primes:
                while self._candidate > i[1]:
                    i[1] = i[0] + i[1]

                if self._candidate == i[1]:
                    break
            else:
                # if here, we have i == primes[-1]:
                self._primes.append([self._candidate, self._candidate])
                return self._candidate


class _CachedPrimesCalculator:  # pylint: disable=too-few-public-methods
    """Utility class to compute and cache the first n prime numbers."""

    def __init__(self) -> None:
        """Initialize the object."""
        self._primes_iterator = _PrimesIterator()
        self._cached_primes: List[int] = [2]

    def get_n_primes(self, n: int) -> NDArray[np.int64]:
        """
        Get the first n primes.

        Args:
            n: the number of primes.

        Returns:
            a list containing the first n primes.
        """
        check_arg(n >= 1, "input must be greater than 0")
        if n <= len(self._cached_primes):
            return np.array(self._cached_primes[:n])

        nb_next_primes = n - len(self._cached_primes)
        next_primes = itertools.islice(self._primes_iterator, nb_next_primes)
        self._cached_primes.extend(next_primes)
        return np.array(self._cached_primes[:n])


def halton(
    sample_size: int, bases: NDArray[np.int64], n_start: int
) -> NDArray[np.float64]:
    """
    Van der Corput sequence, generalized as to accept a starting point in the sequence.

    Args:
        sample_size:  number of element of the sequence
        bases: bases of the sequence
        n_start: starting point of the sequence

    Returns:
        sequence of Halton
    """
    check_arg(sample_size > 0, "sample size must be greater than zero")
    check_arg(bool((bases > 1).all()), "based must be greater than one")
    check_arg(n_start >= 0, "n_start must be greater or equal zero")
    nb_bases = len(bases)
    sequence: NDArray[np.float64] = np.zeros(shape=(sample_size, nb_bases))
    for index in range(n_start + 1, sample_size + n_start + 1):
        n_th_numbers: NDArray[np.float64] = np.zeros(shape=nb_bases)
        denoms: NDArray[np.float64] = np.ones(shape=nb_bases)
        done: NDArray[np.bool8] = np.zeros(shape=nb_bases, dtype=np.bool8)
        i = np.repeat(np.int64(index), repeats=nb_bases)
        while (i > 0).any():
            i, remainders = np.divmod(i, bases)
            denoms *= bases
            # mask reminders in case i = 0
            remainders[done] = 0.0
            n_th_numbers += remainders / denoms
            done[i == 0] = True
        sequence[index - 1 - n_start, :] = n_th_numbers
    return sequence



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['black_it',
 'black_it.loss_functions',
 'black_it.plot',
 'black_it.samplers',
 'black_it.utils']

package_data = \
{'': ['*']}

install_requires = \
['GPy>=1.10.0,<2.0.0',
 'ipywidgets>=7.7.0,<8.0.0',
 'matplotlib>=3.5.2,<4.0.0',
 'numpy>=1.23.3,<1.24.0',
 'pandas>=1.4.2,<2.0.0',
 'scikit-learn>=1.1.0,<2.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'statsmodels>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'black-it',
    'version': '0.1.2',
    'description': 'black-it: Black-box abm calibration kit',
    'long_description': '\n<p align="center">\n<img src="https://raw.githubusercontent.com/bancaditalia/black-it/main/docs/logo/logo_1024.png" width="500">\n<sup><a href="#footnote-1">*</a></sup>\n</p>\n\n<a href="https://pypi.org/project/black-it">\n    <img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/black-it" />\n</a>\n\n<a href="https://github.com/bancaditalia/black-it/blob/main/LICENSE">\n    <img alt="GitHub" src="https://img.shields.io/github/license/bancaditalia/black-it">\n</a>\n\n<a href="https://codecov.io/gh/bancaditalia/black-it">\n  <img src="https://codecov.io/gh/bancaditalia/black-it/branch/main/graph/badge.svg" />\n</a>\n\n# Black-box abm calibration kit\n\nBlack-it is an easy-to-use toolbox designed to help you calibrate the parameters\nin your agent-based models and simulations (ABMs), using state-of-the-art\ntechniques to sample the parameter search space, with no need to reinvent the\nwheel.\n\nModels from economics, epidemiology, biology, logistics, and more can be dealt\nwith. The software can be used as-is - if your main interest is the ABM model\nitself. However, in case your research thing is to, e.g., devise new sampling\nstrategies for ginormous search spaces and highly non-linear model, then you can\ndeploy and test your new ideas on a solid, reusable, modular foundation, in a\nmatter of days, with no need to reimplement all the plumbings from scratch.\n\n## Installation\n\nThis project requires Python v3.8 or later.\n\nTo install the latest version of the package from [PyPI](https://pypi.org/project/black-it/):\n```\npip install black-it\n```\n\nOr, directly from GitHub:\n\n```\npip install git+https://github.com/bancaditalia/black-it.git#egg=black-it\n```\n\nIf you\'d like to contribute to the package, please read the [CONTRIBUTING.md](./CONTRIBUTING.md) guide.\n\n## Quick Example\n\nThe GitHub repo of Black-it contains a series ready-to-run calibration examples.\n\nTo experiment with them, simply clone the repo and enter the `examples` folder\n\n```\ngit clone https://github.com/bancaditalia/black-it.git\ncd black-it/examples\n```\n\nYou\'ll find several scripts and notebooks. The following is the script named `main.py`, note that copying and pasting \nthe lines below will not work in general as the script needs to be inside the "examples" folder in order to run correctly. \n\n```python\nimport models.simple_models as md\n\nfrom black_it.calibrator import Calibrator\nfrom black_it.loss_functions.msm import MethodOfMomentsLoss\nfrom black_it.samplers.best_batch import BestBatchSampler\nfrom black_it.samplers.halton import HaltonSampler\nfrom black_it.samplers.random_forest import RandomForestSampler\n\ntrue_params = [0.20, 0.20, 0.75]\nbounds = [\n    [0.10, 0.10, 0.10],  # LOWER bounds\n    [1.00, 1.00, 1.00],  # UPPER bounds\n]\nbounds_step = [0.01, 0.01, 0.01]  # Step size in range between bounds\n\nbatch_size = 8\nhalton_sampler = HaltonSampler(batch_size=batch_size)\nrandom_forest_sampler = RandomForestSampler(batch_size=batch_size)\nbest_batch_sampler = BestBatchSampler(batch_size=batch_size)\n\n# define a model to be calibrated\nmodel = md.MarkovC_KP\n\n# generate a synthetic dataset to test the calibrator\nN = 2000\nseed = 1\nreal_data = model(true_params, N, seed)\n\n# define a loss\nloss = MethodOfMomentsLoss()\n\n# define the calibration seed\ncalibration_seed = 1\n\n# initialize a Calibrator object\ncal = Calibrator(\n    samplers=[halton_sampler, random_forest_sampler, best_batch_sampler],\n    real_data=real_data,\n    model=model,\n    parameters_bounds=bounds,\n    parameters_precision=bounds_step,\n    ensemble_size=3,\n    loss_function=loss,\n    random_state=calibration_seed,\n)\n\n# calibrate the model\nparams, losses = cal.calibrate(n_batches=5)\n\nprint(f"True parameters:       {true_params}")\nprint(f"Best parameters found: {params[0]}")\n```\n\nWhen the calibration terminates (~half a minute), towards the end  of the output\nyou should see the following messages:\n```\nTrue parameters:       [0.2, 0.2, 0.75]\nBest parameters found: [0.19 0.19 0.74]\n```\n\n## Docs\n\nBlack-it calibration is initiated via the [Calibrator](https://bancaditalia.github.io/black-it/calibrator/) which,\nwhen called, performs three main steps.\n\nFirst, a [Sampler](https://bancaditalia.github.io/black-it/samplers/) is summoned to suggest a set of promising \nparameter configurations to explore.\n\nSecond, the [model](https://bancaditalia.github.io/black-it/simulator_interface/) to be calibrated is simulated for \nall the selected parameters.\n\nThird, a specific [loss function](https://bancaditalia.github.io/black-it/losses/), measuring the goodness of fitness \nof the simulation data with respect to the real data, is evaluated.\n\nThese steps are performed in a loop, and this allows the samplers to progress towards better parameter values \nby exploiting the knowledge of previously computed loss functions.\n\nA more detailed explanation of how Black-it works is available \n[here](https://bancaditalia.github.io/black-it/description/), while the full documentation -complete with examples \nand tutorials- is available [here](https://bancaditalia.github.io/black-it/). \n\n## Disclaimer\n\nThis package is an outcome of a research project. All errors are those of the authors. All views expressed are personal views, not those of Bank of Italy.\n\n---\n\n<p id="footnote-1">\n* Credits to <a href="https://www.bankit.art/people/sara-corbo">Sara Corbo</a> for the logo.\n</p>\n',
    'author': 'Applied Research Team',
    'author_email': 'appliedresearchteam@bancaditalia.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bancaditalia/black-it',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
